/*
 * Written for SGI DMedia API by Masanao Izumo <mo@goice.co.jp>
 * Ported to OSS by David Olofson <do@reologica.se>, 2001
 *
FIXME: The command FIFO I added isn't timestamped.
FIXME: /David
 *
 * TODO:
 *	* Add resampling filter to avoid sound card
 *	  sample rate dependency.
 *
 *	* Add compressor to avoid clipping distortion.
 *
 *	* Add mmap() support for lowlatency sfx.
 */

#define DBG	if(0)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef OSS_SUPPORT
#include <pthread.h>
#include <unistd.h>
#include "audiodev.h"
#endif

#include "SDL.h"
#include "SDL_audio.h"

#include "audio.h"
#include "sfifo.h"

#define BUFFERS		3
#define	MAX_COMMANDS	64

/*----------------------------------------------------------
	Asynchronous command interface stuff
----------------------------------------------------------*/

typedef struct
{
	enum
	{
		CMD_VOLUME = 0,
		CMD_STOP,
		CMD_PLAY,
		CMD_CONTROL
	} action;
	int	cid;
	union
	{
		int	vol;
		struct
		{
			char	*start;
			int	length;
		} sample;
		struct
		{
			int	ctl;
			int	arg;
		} control;
	} args;
} command_t;

static sfifo_t commands;


/*----------------------------------------------------------
	Audio channel
----------------------------------------------------------*/

typedef Sint8 audio_samp_t;
typedef Sint16 out_samp_t;

typedef struct
{
	audio_samp_t *samp;
	int	nsamp;	/* number of samples */
	int	position;
	int	vol;	/* percent of default */
	int	ctl[AC_COUNT];
} audio_channel_t;

static audio_channel_t audio_channel[MAX_CHANNEL];


/*----------------------------------------------------------
	Engine stuff
----------------------------------------------------------*/

#ifdef OSS_SUPPORT
static audiodev_t adev;
static pthread_t engine_thread;
#endif

static int using_oss = 0;

static SDL_AudioSpec audiospec;

static int running = 0;
static int master_vol = 100<<2;

/*
 * Messy. bufsize is in mono samples, while we're
 * actually dealing with stereo here.
 */
static int *mixbuf = NULL;
static int bufsize = 64;


/*----------------------------------------------------------
	Engine code
----------------------------------------------------------*/

static void __play_stop(int cid)
{
	audio_channel_t *c;
	c = audio_channel + cid;
	c->samp = NULL;
}


static void __play(int cid, char *samp, int nsamp)
{
	audio_channel_t *c;
	c = audio_channel + cid;
	c->samp = (audio_samp_t *) samp;
	c->nsamp = nsamp;
	c->position = 0;
	if(c->ctl[AC_PLAYRESET])
		memset(c->ctl+1, 0, AC_COUNT-1);
}


static void __audio_volume(int cid, int vol)
{
	if (cid == -1)
	{
		int i;
		for (i = 0; i < MAX_CHANNEL; i++)
			audio_channel[i].vol = vol;
	}
	else
		audio_channel[cid].vol = vol;
}


static void __run_commands(void)
{
	while(sfifo_used(&commands) >= sizeof(command_t))
	{
		command_t cmd;
		if(sfifo_read(&commands, &cmd, sizeof(cmd)) < 0)
		{
			fprintf(stderr, "audio.c: Engine failure!\n");
			running = 0;
			return;
		}
		switch(cmd.action)
		{
		  case CMD_VOLUME:
			DBG printf("Got CMD_VOLUME.\n");
			__audio_volume(cmd.cid, cmd.args.vol);
			break;
		  case CMD_STOP:
			DBG printf("Got CMD_STOP.\n");
			__play_stop(cmd.cid);
			break;
		  case CMD_PLAY:
			DBG printf("Got CMD_PLAY.\n");
			__play(cmd.cid, cmd.args.sample.start,
					cmd.args.sample.length);
			break;
		  case CMD_CONTROL:
			DBG printf("Got CMD_CONTROL.\n");
			{
				audio_channel_t *c;
				c = audio_channel + cmd.cid;
				c->ctl[cmd.args.control.ctl] =
						cmd.args.control.arg;
			}
			break;
		}

	}
}


static void __push_command(command_t *cmd)
{
	if(!running)
		return;

	if(sfifo_space(&commands) < sizeof(command_t))
	{
		fprintf(stderr, "WARNING: Audio command FIFO overflow!\n");
		return;
	}
	sfifo_write(&commands, cmd, sizeof(command_t));
}


static void __mix(int *buf, int frames)
{
	int i;
	memset(buf, 0, frames * sizeof(int));
	for (i = 0; i < MAX_CHANNEL; i++)
	{
		audio_channel_t *c;
		c = audio_channel + i;

		if(!c->samp)
			continue;
		
		/* Mix from 8 bit source */
		if(c->position < c->nsamp)
		{
			int j, n, v;
			audio_samp_t *s;

			n = c->nsamp - c->position;
			v = c->vol;
			if (n > frames)
				n = frames;
			s = c->samp + c->position;

			if (v == DEFAULT_VOL)
			{
				for (j = 0; j < n; j++)
					buf[j] += s[j]<<8;
			}
			else
			{
				int vm = (double)v / (double)DEFAULT_VOL;
				for (j = 0; j < n; j++)
					buf[j] += s[j] * vm;
			}
			c->position += n;
		}

		/* Handle end-of-samples */
		if(c->position >= c->nsamp)
		{
			if(c->ctl[AC_LOOP])
				c->position = 0;
			else
				c->samp = NULL;
		}
	}
}


/*
 * Apply master volume and convert with saturation
 */
static void __clip(int *inbuf, out_samp_t *outbuf, int frames)
{
	int i;
	for(i = 0; i < frames; ++i)
	{
		int s = (int)inbuf[i];
		s *= master_vol;
		s >>= 8;
		if(s < -32768)
			outbuf[i] = -32768;
		else if(s > 32767)
			outbuf[i] = 32767;
		else
			outbuf[i] = s;
	}
}


/*
 * Engine callback for SDL_audio and OSS
 */
static void __audio_callback(void *ud, Uint8 *stream, int len)
{
	out_samp_t *outbuf = (out_samp_t *)stream;
	int frames = len / sizeof(out_samp_t);
	__run_commands();
	__mix(mixbuf, frames);
	__clip(mixbuf, outbuf, frames);
}


#ifdef OSS_SUPPORT
/*
 * Engine thread for OSS
 */
void *__audio_engine(void *dummy)
{
	int bytes = bufsize * sizeof(out_samp_t);
	out_samp_t *outbuf = malloc(bytes);
	if(!outbuf)
	{
		audiodev_close(&adev);
		running = 0;
		return NULL;
	}
	while(running)
	{
		__audio_callback(NULL, (Uint8 *)outbuf, bytes);
		write(adev.outfd, outbuf, bytes);
	}
	audiodev_close(&adev);
	return NULL;
}
#endif


static int __init_oss_audio()
{
#ifdef OSS_SUPPORT
	mixbuf = malloc(bufsize * sizeof(int));
	if(!mixbuf)
		return -1;

	audiodev_init(&adev);
	adev.rate = SAMPLE_RATE;
	adev.fragmentsize = bufsize * sizeof(out_samp_t);
	adev.fragments = BUFFERS;
	running = 1;
	if(audiodev_open(&adev) < 0)
	{
		sfifo_close(&commands);
		fprintf(stderr, "audio.c: Failed to open audio!\n");
		running = 0;
		return -2;
	}

	if(pthread_create(&engine_thread, NULL, __audio_engine, NULL))
	{
		fprintf(stderr, "audio.c: Failed to start audio engine!\n");
		return -3;
	}
	return 0;
#else
	fprintf(stderr, "OSS audio not compiled in!\n");
#endif
}


static int __init_SDL_audio()
{
	SDL_AudioSpec as;

	if(SDL_InitSubSystem(SDL_INIT_AUDIO) < 0)
		return -2;

	as.freq = SAMPLE_RATE;
	as.format = AUDIO_S16SYS;
	as.channels = 2;
	as.samples = bufsize / as.channels;
	as.callback = __audio_callback;
	if(SDL_OpenAudio(&as, &audiospec) < 0)
		return -3;

	if(audiospec.format != AUDIO_S16SYS)
	{
		SDL_CloseAudio();
		return -4;
	}

	mixbuf = malloc(audiospec.samples *
			sizeof(int) * audiospec.channels);
	if(!mixbuf)
	{
		SDL_CloseAudio();
		return -1;
	}

	running = 1;
	SDL_PauseAudio(0);
	return 0;
}


/*----------------------------------------------------------
	Open/close code
----------------------------------------------------------*/

int audio_open(int latency, int use_oss)
{
	int i;
	running = 0;

	using_oss = use_oss;

	if(sfifo_init(&commands, sizeof(command_t) * MAX_COMMANDS) < 0)
	{
		fprintf(stderr, "audio.c: Failed to set up audio engine!\n");
		return -1;
	}

	memset(audio_channel, 0, sizeof(audio_channel));

	for (i = 0; i < MAX_CHANNEL; i++)
	{
		audio_channel[i].vol = DEFAULT_VOL;
		audio_channel[i].ctl[AC_PLAYRESET] = 1;
	}

	/*
	 * Note: We assume here that SDL also uses 3
	 *       fragments, which may not be true.
	 */
	if(latency > 1000)
		latency = 1000;
	bufsize = 64;
	while(BUFFERS*bufsize*1000 / (SAMPLE_RATE*2) < latency)
		bufsize <<= 1;
	bufsize >>= 1;

	if(using_oss)
		return __init_oss_audio();
	else
		return __init_SDL_audio();
}


void audio_close(void)
{
	if(!running)
		return;

	printf("Stopping audio engine... ");
	running = 0;
	if(using_oss)
#ifdef OSS_SUPPORT
		pthread_join(engine_thread, NULL);
#else
		;
#endif
	else
		SDL_CloseAudio();
	printf("Ok.\n");
}


/*----------------------------------------------------------
	Channel control API code
----------------------------------------------------------*/

void audio_master_volume(int vol)
{
	master_vol = vol<<2;
}


void audio_volume(int cid, int vol)
{
	command_t cmd;
	cmd.action = CMD_VOLUME;
	cmd.cid = cid;
	cmd.args.vol = vol;
	__push_command(&cmd);
}

void audio_play(int cid, char* samp, int nsamp)
{
	command_t cmd;
	cmd.action = CMD_PLAY;
	cmd.cid = cid;
	cmd.args.sample.start = samp;
	cmd.args.sample.length = nsamp;
	__push_command(&cmd);
}

void audio_control(int cid, int ctl, int arg)
{
	command_t cmd;
	cmd.action = CMD_CONTROL;
	cmd.cid = cid;
	cmd.args.control.ctl = ctl;
	cmd.args.control.arg = arg;
	__push_command(&cmd);
}

void audio_stop(int cid)
{
	command_t cmd;
	if(cid < 0)
		return;

	cmd.action = CMD_STOP;
	cmd.cid = cid;
	__push_command(&cmd);
}


/*
 * audio_is_active() bypasses the lock-free FIFO.
 * OTOH, it's meant to be asynchronous (I think...),
 * so it should do the right thing, provided the
 * reads are atomic.
 *		/David
 */
int audio_is_active(int cid)
{
	audio_channel_t *c;
	if(!running)
		return 0;

	c = audio_channel + cid;
	return c->samp != NULL;
}
